import struct
import time
import cmath
import numpy as np

from qulacs import QuantumState, QuantumCircuit
from qulacs.gate import StateReflection
from qulacs.gate import X, Z, H, SWAP, RZ, RY, to_matrix_gate, Measurement, DenseMatrix, SparseMatrix, CNOT
from qulacs.gate import X as X_gate

significant_dig = 6 # no. of energy register qubits (n_reg); 6 or 8 for acceptable cost

np.random.seed(seed=32)

def to_binay_array(num, m):
    """Convert a positive integer num into an m-bit bit vector"""
    return np.array(list(np.binary_repr(num).zfill(m))).astype(np.int8)

def float_to_bin(f):
    if (f == 0.0):
        return (0, 0, "0"*53)
    b = (bin(struct.unpack('>Q', struct.pack('>d', f))[0])[2:]).zfill(64)
    sign = int(b[0])
    exponent = int(b[1:12],2)-1023
    fraction = '1'+b[12:]
    return (sign, exponent, fraction)

def dig_cut(f, m):
    (sign, exponent, fraction) = float_to_bin(f)
    effect_dig = max(exponent+1+m,0)
    if effect_dig == 0:
        return 0.0
    if sign == 0:
        f  = int(fraction[:effect_dig],2)*2**(exponent-effect_dig+1)
    else:
        f  = -int(fraction[:effect_dig],2)*2**(exponent-effect_dig+1)
    return f

def R_theta(target,angle): # R_theta = [[1, 0],[0,e^iθ]]
    return DenseMatrix(target, np.array( [[1,0], [0,np.cos(angle)+1.j*np.sin(angle)]]))

def R_2pi_phase(target,angle): # R_theta = [[e^2πiθ, 0][0,e^2πiθ]
    angle *= 2*np.pi
    return DenseMatrix(target, np.array( [[np.cos(angle)+1.j*np.sin(angle),0], [0,np.cos(angle)+1.j*np.sin(angle)]]))

def inverse_qft(state, start, m):
    # for swap
    for qft_target in range(m//2):
        target1 = start + qft_target
        target2 = start + (m-1-qft_target)
        SWAP(target1,target2).update_quantum_state(state)

    for qft_target in range(m):
        target = start + qft_target
        for rotate_target in range(qft_target, 0, -1):
            R_k_inverse = R_theta(target, -2*np.pi/(2**(rotate_target+1)))
            R_k_inverse.add_control_qubit(target-rotate_target, 1)
            R_k_inverse.update_quantum_state(state)
        H(target).update_quantum_state(state)
    return state

def qft(state, start, m):
    for qft_target in range(m-1, -1, -1):
        target = start + qft_target
        H(target).update_quantum_state(state)
        for control_target in range(qft_target-1,-1,-1):
            R_k = R_theta(target, 2*np.pi/(2**(qft_target-control_target+1)))
            R_k.add_control_qubit(control_target+start, 1)
            R_k.update_quantum_state(state)
    # for swap
    for qft_target in range(m//2):
        target1 = start + qft_target
        target2 = start + (m-1-qft_target)
        SWAP(target1,target2).update_quantum_state(state)
    return state


def U_a(target, phi):
    return R_theta(target, 2*np.pi*phi)

def U_w(target, control, phi):
    if target == control:
        return U_a(target, phi)
    u = R_theta(target, 2*np.pi*phi)
    u.add_control_qubit(control, 1)
    return u

def Uw(anc, m):
    Uw = DenseMatrix(anc, np.array( [[-1,0], [0,1]]))
    for k in range(anc + 1, anc+m):
        Uw.add_control_qubit(k, 0)
    return Uw

def Us(state):
    Us = StateReflection(state.copy())
    return Us

def A_gate(i, j, theta, state):
    CNOT(j,i).update_quantum_state(state)
    RY(j, -(theta+np.pi/2.0)).update_quantum_state(state)
    CNOT(i,j).update_quantum_state(state)
    RY(j, (theta+np.pi/2.0)).update_quantum_state(state)
    CNOT(j,i).update_quantum_state(state)
    return state

def number_conserve(state):
    X_gate(0).update_quantum_state(state)
    X_gate(2).update_quantum_state(state)

    RY(4, -2*np.arcsin(np.sqrt(1/6))).update_quantum_state(state)
    CNOT(4,6).update_quantum_state(state)
    CNOT(4,0).update_quantum_state(state)
    CNOT(6,2).update_quantum_state(state)

    mat = to_matrix_gate(RY(4, -2*np.arcsin(np.sqrt(4/5))))
    mat.add_control_qubit(2, 1)
    mat.update_quantum_state(state)
    mat = to_matrix_gate(X_gate(0))
    mat.add_control_qubit(2, 1)
    mat.add_control_qubit(4, 1)
    mat.update_quantum_state(state)

    state = A_gate(0,2,np.pi*5/4, state)
    state = A_gate(4,6,-np.pi/4, state)
    ##
    X_gate(1).update_quantum_state(state)
    X_gate(3).update_quantum_state(state)

    RY(5, -2*np.arcsin(np.sqrt(1/6))).update_quantum_state(state)
    CNOT(5,7).update_quantum_state(state)
    CNOT(5,1).update_quantum_state(state)
    CNOT(7,3).update_quantum_state(state)

    mat = to_matrix_gate(RY(5, -2*np.arcsin(np.sqrt(4/5))))
    mat.add_control_qubit(3, 1)
    mat.update_quantum_state(state)
    mat = to_matrix_gate(X_gate(1))
    mat.add_control_qubit(3, 1)
    mat.add_control_qubit(5, 1)
    mat.update_quantum_state(state)

    state = A_gate(1,3,np.pi*5/4, state)
    state = A_gate(5,7,-np.pi/4, state)

def Sample(state, b, w):
    n = len(b)
    m = significant_dig
    assert w.shape == (n, n), "b, w has invalid shapes"
    assert state.get_qubit_count() == n+m+m, "state is not n+m+m qubit"

    a = np.zeros(n)
    W = np.zeros((n,n))
    for i in range(n):
        a[i] = b[i]
    for i in range(n):
        for j in range(n):
            if i == j:
                a[i] += w[i,j]
            elif i < j:
                W[i,j] += w[i,j]
            elif i > j:
                W[j,i] += w[i,j] # 左下は反転

    E_exact = np.zeros(2**n)
    E_max = -np.inf
    E_min = np.inf
    for i in range(2**n):
        i_array = to_binay_array(i, n)[::-1]
        if (sum(i_array[0::2]) != n//4) or (sum(i_array[1::2]) != n//4):
            continue
        for k in range(n):
            E_exact[i] += a[k] * i_array[k]
        for k in range(n):
            for l in range(k+1,n):
                E_exact[i] += W[k,l]*i_array[k]*i_array[l]
        E_max = max(E_max, E_exact[i])
        E_min = min(E_min, E_exact[i])

    exact_list = np.exp(E_exact/2)
    for i in range(2**n):
        i_array = to_binay_array(i, n)[::-1]
        if (sum(i_array[0::2]) != n//4) or (sum(i_array[1::2]) != n//4):
            exact_list[i] = 0.0
    norm = np.sqrt(sum(exact_list**2))

    exact_list /= norm
    # 1-2**(-significant_dig) = (E_max-E_min)/width
    # width = 2**(significant_dig+1)*(E_max - E_min)/(2**(significant_dig+1)-1)
    for z in range(1,n+n*n+1):
        width = (E_max - E_min)/(1-z*2**(-significant_dig)+2**(-significant_dig-10))
        #print('z:', z)
        #print('width:', width)

        a_shift = a/width
        w_shift = W/width

        for i in range(n):
            a_shift[i] = dig_cut(a_shift[i],m)
        for i in range(n):
            for j in range(i+1,n):
                w_shift[i,j]= dig_cut(w_shift[i,j], m)
        x = 0
        i = np.argmax(E_exact)
        i_array = to_binay_array(i, n)[::-1]
        for k in range(n):
            x += a_shift[k] * i_array[k]
        for k in range(n):
            for l in range(k+1,n):
                x += w_shift[k][l]*i_array[k]*i_array[l]
        E_shift = -(1-2**(-significant_dig)-x)
        E_shift = dig_cut(E_shift,m)
        e = np.zeros(2**n)
        e_max = -np.inf
        e_min = np.inf
        for i in range(2**n):
            i_array = to_binay_array(i, n)[::-1]
            if (sum(i_array[0::2]) != n//4) or (sum(i_array[1::2]) != n//4):
                continue
            e[i] -= E_shift
            i_array = to_binay_array(i, n)[::-1]
            for k in range(n):
                e[i] += a_shift[k] * i_array[k]
            for k in range(n):
                for l in range(k+1,n):
                    e[i] += w_shift[k][l]*i_array[k]*i_array[l]
            e_max = max(e_max, e[i])
            e_min = min(e_min, e[i])
        #e_max = max(e)
        # print("max(e):", e_max)
        #e_min = min(e)
        # print("min(e):", min(e))
        if (e_max <= 1-2**(-m) and e_min >= 0):
            break
    if z == n+n*n:
        print("z is unnatural")
        exit(1)
    print("width:", width)
    x = 0
    i = np.argmax(E_exact)
    i_array = to_binay_array(i, n)[::-1]
    for k in range(n):
        x += a_shift[k] * i_array[k]
    for k in range(n):
        for l in range(k+1,n):
            x += w_shift[k][l]*i_array[k]*i_array[l]
    x -= E_shift
    # print("E_max")
    # print(x)
    # print(float_to_bin(x))
    e_list = np.exp(e*width/2)
    for i in range(2**n):
        i_array = to_binay_array(i, n)[::-1]
        if (sum(i_array[0::2]) != n//4) or (sum(i_array[1::2]) != n//4):
            e_list[i] = 0
    norm = np.sqrt(sum(e_list**2))
    e_list /= norm

    # print("exact")
    # print(exact_list)
    # print("approx")
    # print(e_list)
    # print("diff")
    # print(exact_list-e_list)

    sign, max_exp, fraction = float_to_bin(x)

    st = time.time()
    # start gate
    number_conserve(state)
    #hadmard# for i in range(n):
    #hadmard#     H(i).update_quantum_state(state)
    for j in range(m):
        H(n+j).update_quantum_state(state)

    for j in range(m):
        r = 2**j
        for i in range(n):
            U_r = U_a(i,r*a_shift[i])
            U_r.add_control_qubit(n+j, 1)
            U_r.update_quantum_state(state)

        for i1 in range(n):
            for i2 in range(i1+1,n):
                U_r = U_w(i1,i2,r*w_shift[i1,i2])
                U_r.add_control_qubit(n+j, 1)
                U_r.update_quantum_state(state)
        U_r = R_2pi_phase(0, r*(-E_shift))
        U_r.add_control_qubit(n+j,1)
        U_r.update_quantum_state(state)

    # inverse qft
    state = inverse_qft(state, n, m)
    # for swap
    count = 0
    for bit in range(m):
        exponent = max_exp - bit
        controlled_Ry = to_matrix_gate(RY(n+m+bit, 2*np.arccos(np.exp(-width/2*2**exponent)))) # /2 is for sqrt
        controlled_Ry.add_control_qubit(n+(m-1-bit), 0)
        controlled_Ry.update_quantum_state(state)
    state = qft(state, n, m)
    for j in range(m-1,-1,-1):
        r = 2**j
        for i in range(n):
            U_r = U_a(i,-r*a_shift[i])
            U_r.add_control_qubit(n+j, 1)
            U_r.update_quantum_state(state)
        for i1 in range(n):
            for i2 in range(i1+1, n):
                U_r = U_w(i1,i2,-r*w_shift[i1,i2])
                U_r.add_control_qubit(n+j, 1)
                U_r.update_quantum_state(state)
        U_r = R_2pi_phase(0, r*(E_shift))
        U_r.add_control_qubit(n+j,1)
        U_r.update_quantum_state(state)
    for j in range(m-1,-1,-1):
        H(n+j).update_quantum_state(state)

    marginal_list = [2]*(n+m)+[0]*m
    marginal_probability = state.get_marginal_probability(marginal_list)
    print("marginal_prob:", marginal_probability)

    theta_prob = np.arcsin(np.sqrt(min(marginal_probability,1)))
    best_k = ((0.5 * np.pi / theta_prob) - 1) / 2.0
    best_k_round = np.round(best_k).astype(np.int64)

    uw = Uw(n+m, m)
    us = Us(state)
    for tmp in range(best_k_round):
        uw.update_quantum_state(state)
        us.update_quantum_state(state)
    marginal_probability = state.get_marginal_probability(marginal_list)
    print("amplified {0} time prob get 0: {1}".format(best_k_round,marginal_probability))

    M_list = [Measurement(n+m+i, n+m+i) for i in range(m)]
    sub_state = state.copy()
    result = 1
    count = 0
    while(count < 1000):
        count += 1
        flag = True
        for i in range(m):
            M_list[i].update_quantum_state(state)
            result = state.get_classical_value(n+m+i)
            if result == 1:
                flag = False
                break
        if flag:
            break
        else:
            state = sub_state.copy()
    del sub_state
    et = time.time()
    # print("mesurement time:", count)
    print("prepare time:",et-st,"[s]")
    l = state.get_vector()[:2**n]
    state = QuantumState(n)
    state.load(l)
    return state

def PhaseGate(target, angle):
    return DenseMatrix(target, np.array( [[1,0], [0,np.cos(angle)+1.j*np.sin(angle)]]))

def CPhaseGate(target, control, angle):
    """
    Create controlled phase gate diag(1,e^{i*angle}) with controll. (Qulacs.gate is requried)

    Args:
      int target:  index of target qubit.
      int control:  index of control qubit.
      float64 angle: angle of phase gate.
    Returns:
      QuantumGateBase.DenseMatrix: diag(1, exp(i*angle)).
    """
    if target == control:
        return PhaseGate(target, angle)

    CPhaseGate = DenseMatrix([target], np.array( [[1,0], [0,np.cos(angle)+1.j*np.sin(angle)]])  )

    CPhaseGate.add_control_qubit(control, 1)
    return CPhaseGate

def Phase(state, c, X):
    n = len(c)
    assert X.shape == (n, n) ,"c, X has invalid shapes"

    for i in range(n):
        PhaseGate(i, c[i]/2).update_quantum_state(state)
    for i in range(n):
        for j in range(n):
            CPhaseGate(i, j, X[i,j]/2).update_quantum_state(state)
    return state
