import time
import numpy as np

from qulacs import QuantumState,PauliOperator,Observable,GeneralQuantumOperator
from qulacs.gate import X,Z,RY,Measurement,to_matrix_gate
from qulacs.quantum_operator import create_quantum_operator_from_openfermion_file

class all_expectation:
    """
    state
    Zistate # n list
    ZiZjstate # n x n list
    ZiZjZkstate # n x n x n list

    H
    HZi   # n list
    ZiH   # n list
    HZiZj # n x n list
    ZiZjH # n x n list
    ZiHZj # n x n list
    HZiZjZk # n x n x n list
    ZiHZkZl # n x n x n list
    H
    HZiZjZkZl # n x n x n x n list

    Zi     # n list
    ZiZj   # n x n list
    ZiZjZk # n x n x n list
    """

    def __init__(self, state_, n_, Hamiltonian_, can_hess_, can_SR_):
        st = time.time()
        self.state = state_
        self.qubit = self.state.get_qubit_count()
        self.Hamiltonian = GeneralQuantumOperator(self.qubit)
        for i in range(Hamiltonian_.get_term_count()):
            self.Hamiltonian.add_operator(Hamiltonian_.get_term(i))

        self.can_hess = can_hess_
        self.can_SR = can_SR_
        self.bool_jac = self.bool_SR = self.bool_hess = self.bool_initialize = False
        self.n = n = n_
        self.Zistate = np.array([state_ for i in range(n)])
        self.ZiZjstate = np.array([state_ for _ in range(n*n)])
        self.ZiZjstate = np.resize(self.ZiZjstate,(n,n))

        self.Zi = np.zeros(n, dtype=complex)
        self.vi_2 = np.zeros(n, dtype=complex)
        self.HZi = np.zeros(n, dtype=complex)
        self.ZiZj = np.zeros((n, n), dtype=complex)
        self.HZiZj = np.zeros((n, n), dtype=complex)

        if can_SR_ or can_hess_:
            self.ZiZjZk = np.zeros((n, n, n), dtype=complex)
            self.ZiZjZkZl = np.zeros((n, n, n, n), dtype=complex)

        if can_hess_:
            self.ZiZjZkstate = np.array([state_ for _ in range(n*n*n)])
            self.ZiZjZkstate = np.resize(self.ZiZjZkstate,(n,n,n))
            self.ZiZjZkZlstate = np.array([state_ for _ in range(n*n*n*n)])
            self.ZiZjZkZlstate = np.resize(self.ZiZjZkZlstate,(n,n,n,n))

            self.ZiHZj = np.zeros((n, n), dtype=complex) #
            self.ZiH = np.zeros(n, dtype=complex) #
            self.ZiZjH = np.zeros((n, n), dtype=complex) #
            self.HZiZjZk = np.zeros((n, n, n), dtype=complex) #
            self.ZiHZjZk = np.zeros((n, n, n), dtype=complex) #
            self.ZiZjHZk = np.zeros((n, n, n), dtype=complex) #
            self.HZiZjZkZl = np.zeros((n, n, n, n), dtype=complex)
            self.ZiZjHZkZl = np.zeros((n, n, n, n), dtype=complex)

        for i in range(n):
            self.Zistate[i] = state_.copy()
            Z(i).update_quantum_state(self.Zistate[i])
            for j in range(n):
                self.ZiZjstate[i,j] = self.Zistate[i].copy()
                Z(j).update_quantum_state(self.ZiZjstate[i,j])

                if can_hess_:
                    for k in range(n):
                        self.ZiZjZkstate[i,j,k] = self.ZiZjstate[i,j].copy()
                        Z(k).update_quantum_state(self.ZiZjZkstate[i,j,k])
                        for l in range(n):
                            self.ZiZjZkZlstate[i,j,k,l] = self.ZiZjZkstate[i,j,k].copy()
                            Z(l).update_quantum_state(self.ZiZjZkZlstate[i,j,k,l])

        #print("finish time:", time.time()-st)

    def initialize(self):
        st = time.time()
        self.H = self.Hamiltonian.get_expectation_value(self.state)
        n = self.n
        Hami = self.Hamiltonian
        #print("Energy", self.H.real)
        I = PauliOperator(f" ", 1.0)
        for i in range(self.n):
            self.Zi[i] = I.get_transition_amplitude(self.state, self.Zistate[i])
            self.HZi[i] = Hami.get_transition_amplitude(self.state, self.Zistate[i])
            for j in range(self.n):
                self.ZiZj[i,j] = I.get_transition_amplitude(self.state, self.ZiZjstate[i,j])
                self.HZiZj[i,j] = Hami.get_transition_amplitude(self.state, self.ZiZjstate[i,j])
                if self.can_SR or self.can_hess:
                    for k in range(self.n):
                        self.ZiZjZk[i,j,k] = I.get_transition_amplitude(self.Zistate[i], self.ZiZjstate[j,k])
                        for l in range(self.n):
                            self.ZiZjZkZl[i,j,k,l] = I.get_transition_amplitude(self.ZiZjstate[i,j], self.ZiZjstate[k,l])
        if self.can_SR or self.can_hess:
            self.vi = (1.0/2)*(1.0-self.Zi)
            self.vivj = np.zeros((n,n), dtype=complex)
            self.vivjvk = np.zeros((n,n,n), dtype=complex)
            self.vivjvkvl = np.zeros((n,n,n,n), dtype=complex)
            for i in range(n):
                for j in range(n):
                    self.vivj[i,j] = (1.0/4)*(1.0-self.Zi[i]-self.Zi[j]+self.ZiZj[i,j])
                    for k in range(n):
                        self.vivjvk[i,j,k] = (1.0/8)*(1.0-self.Zi[i]-self.Zi[j]-self.Zi[k]+self.ZiZj[i,j]+self.ZiZj[j,k]+self.ZiZj[i,k]-self.ZiZjZk[i,j,k])
                        for l in range(n):
                            self.vivjvkvl[i,j,k,l] = (1.0/16)*(1.0-self.Zi[i]-self.Zi[j]-self.Zi[k]-self.Zi[l]+self.ZiZj[i,j]+self.ZiZj[i,k]+self.ZiZj[i,l]+self.ZiZj[j,k]+self.ZiZj[j,l]+self.ZiZj[k,l]-self.ZiZjZk[i,j,k]-self.ZiZjZk[i,j,l]-self.ZiZjZk[i,k,l]-self.ZiZjZk[j,k,l]+self.ZiZjZkZl[i,j,k,l])

        if self.can_hess:
            for i in range(n):
                self.ZiH[i] = Hami.get_transition_amplitude(self.Zistate[i], self.state)
                for j in range(n):
                    self.ZiHZj[i,j] = Hami.get_transition_amplitude(self.Zistate[i], self.Zistate[j])
                    self.ZiZjH[i,j] = Hami.get_transition_amplitude(self.ZiZjstate[i,j], self.state)
                    for k in range(n):
                        self.HZiZjZk[i,j,k] = Hami.get_transition_amplitude(self.state, self.ZiZjZkstate[i,j,k])
                        self.ZiHZjZk[i,j,k] = Hami.get_transition_amplitude(self.Zistate[i], self.ZiZjstate[j,k])
                        self.ZiZjHZk[i,j,k] = Hami.get_transition_amplitude(self.ZiZjstate[i,j], self.Zistate[k])
                        for l in range(n):
                            self.HZiZjZkZl[i,j,k,l] = Hami.get_transition_amplitude(self.state, self.ZiZjZkZlstate[i,j,k,l])
                            self.ZiZjHZkZl[i,j,k,l] = Hami.get_transition_amplitude(self.ZiZjstate[i,j], self.ZiZjstate[k,l])

        #print("initiate_time:", time.time() - st)
        self.bool_initialize = True

    def calculate_jacobian(self):
        assert self.bool_initialize, "Please initialize brfore"
        st = time.time()
        n = self.n
        self.jac = np.zeros(2*n+2*n*n, dtype=float)

        for I in range(n):
            self.jac[I] = self.dH_a1(I)
        for I in range(n, n+n*n):
            i = (I - n)//n
            j = (I-n)%n
            self.jac[I] = self.dH_w1(i,j)
        for I in range(n+n*n, 2*n+n*n):
            i = (I- n-n*n)
            self.jac[I] = self.dH_c1(i)
        for I in range(2*n+n*n, 2*n+2*n*n):
            i = (I-2*n-n*n)//n
            j = (I-2*n-n*n)%n
            self.jac[I] = self.dH_x1(i,j)

        #print("finish jacobian", time.time()-st)
        self.bool_jac = True

    def calculate_Smat(self):
        assert self.bool_initialize, "Please initialize brfore"
        assert self.can_SR, "Cannot calculate Smatrix"

        st = time.time()
        n = self.n
        self.Smat = np.zeros((n+n*n, n+n*n), dtype=float)
        for I in range(n):
            for J in range(I, n+n*n):
                if J < n:
                    self.Smat[I,J] = self.SR_aa(I,J)
                else:
                    k = (J-n)//n
                    l = (J-n)%n
                    self.Smat[I,J] = self.SR_aw(I, k, l)
        for I in range(n, n+n*n):
            for J in range(I, n+n*n):
                i = (I-n)//n
                j = (I-n)%n
                k = (J-n)//n
                l = (J-n)%n
                self.Smat[I,J] = self.SR_ww(i, j, k, l)
        self.Smat = self.Smat+self.Smat.T
        for I in range(n+n*n):
            self.Smat[I,I] /= 2.0
        for I in range(n+n*n):
            self.Smat[I,I] *= (1+1e-4)
        #print("finish calcS", time.time()-st)
        self.bool_SR = True

    def calculate_SRjac(self):
        st = time.time()
        assert self.bool_initialize, "Please initialize brfore"
        assert self.bool_jac, "Please calculate jac before"
        assert self.bool_SR, "Please calculate Smatrix"
        n = self.n
        self.SRjac = np.zeros(2*n+2*n*n, dtype=float)
        self.original_norm_the = np.linalg.norm(self.jac[:n+n*n])
        print("original_norm_the", np.linalg.norm(self.jac[:n+n*n]))
        self.SRjac[:n+n*n] = np.linalg.lstsq(self.Smat, self.jac[:n+n*n], rcond=5e-7)[0]
        self.original_norm_tau = np.linalg.norm(self.jac[n+n*n:])
        print("original_norm_tau", np.linalg.norm(self.jac[n+n*n:]))
        self.SRjac[n+n*n:] = np.linalg.lstsq(self.Smat, self.jac[n+n*n:], rcond=5e-7)[0]

        #print("finish SRjac", time.time()-st)

    def calculate_hess(self):
        assert self.bool_initialize, "Please initialize brfore"
        assert self.can_hess, "Cannot calculate hessian"
        n = self.n
        N = 2*n+2*n*n
        self.hess = np.zeros((2*n+2*n*n, 2*n+2*n*n), dtype=float)
        for I in range(n):
            for J in range(I, N):
                if J < n:
                    self.hess[I,J] = self.Hess_aa(I, J)
                elif J < n+n*n:
                    k = (J-n)//n
                    l = (J-n)%n
                    self.hess[I,J] = self.Hess_aw(I, k, l)
                elif J < 2*n+n*n:
                    j = J-n-n*n
                    self.hess[I,J] = self.Hess_ac(I, j)
                else:
                    k = (J-2*n-n*n)//n
                    l = (J-2*n-n*n)%n
                    self.hess[I,J] = self.Hess_ax(I, k, l)

        for I in range(n, n+n*n):
            for J in range(I, N):
                i = (I-n)//n
                j = (I-n)%n
                if J < n+n*n:
                    k = (J-n)//n
                    l = (J-n)%n
                    self.hess[I,J] = self.Hess_ww(i, j, k, l)
                elif J < 2*n++n*n:
                    k = J-n-n*n
                    self.hess[I,J] = self.Hess_wc(i, j, k)
                else:
                    k = (J-2*n-n*n)//n
                    l = (J-2*n-n*n)%n
                    self.hess[I,J] = self.Hess_wx(i, j, k, l)

        for I in range(n+n*n, 2*n+n*n):
            for J in range(I, N):
                i = I-n-n*n
                if J < 2*n++n*n:
                    j = J-n-n*n
                    self.hess[I,J] = self.Hess_cc(i, j)
                else:
                    k = (J-2*n-n*n)//n
                    l = (J-2*n-n*n)%n
                    self.hess[I,J] = self.Hess_cx(i, k, l)

        for I in range(2*n++n*n, N):
            for J in range(I, N):
                i = (I-2*n-n*n)//n
                j = (I-2*n-n*n)%n
                k = (J-2*n-n*n)//n
                l = (J-2*n-n*n)%n
                self.hess[I,J] = self.Hess_xx(i, j, k, l)

        for I in range(N):
            for J in range(N):
                if I > J:
                    self.hess[I,J] = self.hess[J,I]
        self.bool_hess = True

    def dH_a1(self,i):
        tmp = 0.0
        c = 0.5

        tmp += c*0.5*(self.H - self.HZi[i])
        tmp -= c*self.H*(0.5*(1.0-self.Zi[i]))
        ans = 2*tmp.real
        return ans

    def dH_w1(self, i, j):
        tmp = 0.0
        c = 0.5
        tmp += c*0.25*(self.H-self.HZi[i]-self.HZi[j]+self.HZiZj[i,j])
        tmp -= c*self.H*0.25*(1.0-self.Zi[i]-self.Zi[j]+self.ZiZj[i,j])
        ans = 2*tmp.real
        return ans

    def dH_c1(self, i):
        ans = 0.0
        c = 1.0j/2
        ans = c*0.5*(self.H-self.HZi[i])
        ans = 2*ans.real
        return ans

    def dH_x1(self, i, j):
        ans = 0.0
        c = 1.0j/2
        ans = c*0.25*(self.H-self.HZi[i]-self.HZi[j]+self.HZiZj[i,j])
        ans = 2*ans.real
        return ans

    def SR_aa(self, i, j):
        ans = 0.0
        ans += self.vivj[i,j]/4-self.vi[i]/2*self.vi[j]/2
        return ans.real

    def SR_aw(self, i, k, l):
        ans = 0.0
        ans += self.vivjvk[i,k,l]/4-self.vi[i]/2*self.vivj[k,l]/2
        return ans.real

    def SR_ww(self, i, j, k, l):
        ans = 0.0
        ans += self.vivjvkvl[i,j,k,l]/4-self.vivj[i,j]/2*self.vivj[k,l]/2
        return ans.real

    def Hess_aa(self, i, j):
        ans = 0.0
        ans += (1.0/16)*(self.H-self.HZi[i]-self.HZi[j]+self.HZiZj[i,j])
        ans += -2*((0.25*(self.H-self.HZi[i])).real)*self.vi[j]/2
        ans += -2*(0.25*(self.H-self.HZi[j]))*self.vi[i]/2
        ans += 4*self.H*self.vi[i]/2*self.vi[j]/2
        ans += -2*self.H*self.vivj[i,j]/4
        ans += (1.0/16)*(self.H-self.HZi[j]-self.ZiH[i]+self.ZiHZj[i,j])
        ans = 2*ans.real
        return ans

    def Hess_aw(self, i, k, l):
        ans = 0.0
        ans += (1.0/32)*(self.H-self.HZi[l]-self.HZi[k]+self.HZiZj[k,l]-self.HZi[i]+self.HZiZj[i,l]+self.HZiZj[i,k]-self.HZiZjZk[i,k,l])
        ans += -2*(((1.0/4)*(self.H-self.HZi[i])).real)*self.vivj[k,l]/2
        ans += -2*((1.0/8)*(self.H-self.HZi[k]-self.HZi[l]+self.HZiZj[k,l]))*self.vi[i]/2
        ans += 4*self.H*self.vi[i]/2*self.vivj[k,l]/2
        ans += -2.0*self.H*self.vivjvk[i,k,l]/4
        ans += (1.0/32)*(self.H-self.HZi[l]-self.HZi[k]+self.HZiZj[k,l]-self.ZiH[i]+self.ZiHZj[i,l]+self.ZiHZj[i,k]-self.ZiHZjZk[i,k,l])
        ans = 2*ans.real
        return ans

    def Hess_ac(self, i, j):
        ans = 0.0
        ans += 1.j*(1.0/16)*(self.H-self.HZi[i]-self.HZi[j]+self.HZiZj[i,j])
        ans += -2.j*(0.25*(self.H-self.HZi[j]))*self.vi[i]/2
        ans += 1.j*(1.0/16)*(self.H-self.HZi[j]-self.ZiH[i]+self.ZiHZj[i,j])
        ans = 2*ans.real
        return ans

    def Hess_ax(self, i, k, l):
        ans = 0.0
        ans += 1.j*(1.0/32)*(self.H-self.HZi[l]-self.HZi[k]+self.HZiZj[k,l]-self.HZi[i]+self.HZiZj[i,l]+self.HZiZj[i,k]-self.HZiZjZk[i,k,l])
        ans += -2.j*((1.0/8)*(self.H-self.HZi[k]-self.HZi[l]+self.HZiZj[k,l]))*self.vi[i]/2
        ans += 1.j*(1.0/32)*(self.H-self.HZi[l]-self.HZi[k]+self.HZiZj[k,l]-self.ZiH[i]+self.ZiHZj[i,l]+self.ZiHZj[i,k]-self.ZiHZjZk[i,k,l])

        ans = 2*ans.real
        return ans

    def Hess_ww(self, i, j, k, l):
        ans = 0.0
        ans += (1.0/64)*(self.H-self.HZi[i]-self.HZi[j]-self.HZi[k]-self.HZi[l]+self.HZiZj[i,j]+self.HZiZj[i,k]+self.HZiZj[i,l]+self.HZiZj[j,k]+self.HZiZj[j,l]+self.HZiZj[k,l]-self.HZiZjZk[i,j,k]-self.HZiZjZk[i,j,l]-self.HZiZjZk[i,k,l]-self.HZiZjZk[j,k,l]+self.HZiZjZkZl[i,j,k,l])
        ans += -2*(((1.0/8)*(self.H-self.HZi[i]-self.HZi[j]+self.HZiZj[i,j])).real)*self.vivj[k,l]/2
        ans += -2*((1.0/8)*(self.H-self.HZi[k]-self.HZi[l]+self.HZiZj[k,l]))*self.vivj[i,j]/2
        ans += 4*self.H*self.vivj[i,j]/2*self.vivj[k,l]/2
        ans += -2*self.H*self.vivjvkvl[i,j,k,l]/4
        ans += (1.0/64)*(self.H-self.ZiH[i]-self.ZiH[j]-self.HZi[k]-self.HZi[l]+self.ZiZjH[i,j]+self.ZiHZj[i,k]+self.ZiHZj[i,l]+self.ZiHZj[j,k]+self.ZiHZj[j,l]+self.HZiZj[k,l]-self.ZiZjHZk[i,j,k]-self.ZiZjHZk[i,j,l]-self.ZiHZjZk[i,k,l]-self.ZiHZjZk[j,k,l]+self.ZiZjHZkZl[i,j,k,l])

        ans = 2*ans.real

        return ans

    def Hess_wc(self, i, j, k):
        ans = 0.0
        ans += 1.j*(1.0/32)*(self.H-self.HZi[i]-self.HZi[j]-self.HZi[k]+self.HZiZj[i,j]+self.HZiZj[i,k]+self.HZiZj[j,k]-self.HZiZjZk[i,j,k])
        ans += -2.j*(0.25*(self.H-self.HZi[k]))*self.vivj[i,j]/2
        ans += 1.j*(1.0/32)*(self.H-self.ZiH[i]-self.ZiH[j]-self.HZi[k]+self.ZiZjH[i,j]+self.ZiHZj[i,k]+self.ZiHZj[j,k]-self.ZiZjHZk[i,j,k])
        ans = 2*ans.real
        return ans

    def Hess_wx(self, i, j, k, l):
        ans = 0.0
        ans += 1.j*(1.0/64)*(self.H-self.HZi[i]-self.HZi[j]-self.HZi[k]-self.HZi[l]+self.HZiZj[i,j]+self.HZiZj[i,k]+self.HZiZj[i,l]+self.HZiZj[j,k]+self.HZiZj[j,l]+self.HZiZj[k,l]-self.HZiZjZk[i,j,k]-self.HZiZjZk[i,j,l]-self.HZiZjZk[i,k,l]-self.HZiZjZk[j,k,l]+self.HZiZjZkZl[i,j,k,l])
        ans += -2.j*((1.0/8)*(self.H-self.HZi[k]-self.HZi[l]+self.HZiZj[k,l]))*self.vivj[i,j]/2
        ans += 1.j*(1.0/64)*(self.H-self.ZiH[i]-self.ZiH[j]-self.HZi[k]-self.HZi[l]+self.ZiZjH[i,j]+self.ZiHZj[i,k]+self.ZiHZj[i,l]+self.ZiHZj[j,k]+self.ZiHZj[j,l]+self.HZiZj[k,l]-self.ZiZjHZk[i,j,k]-self.ZiZjHZk[i,j,l]-self.ZiHZjZk[i,k,l]-self.ZiHZjZk[j,k,l]+self.ZiZjHZkZl[i,j,k,l])

        ans = 2*ans.real
        return ans

    def Hess_cc(self, i, j):
        ans = 0.0
        ans += -(1.0/16)*(self.H-self.HZi[i]-self.HZi[j]+self.HZiZj[i,j])
        ans += (1.0/16)*(self.H-self.HZi[j]-self.ZiH[i]+self.ZiHZj[i,j])

        ans = 2*ans.real
        return ans

    def Hess_cx(self, i, k, l):
        ans = 0.0
        ans += -(1.0/32)*(self.H-self.HZi[l]-self.HZi[k]+self.HZiZj[k,l]-self.HZi[i]+self.HZiZj[i,l]+self.HZiZj[i,k]-self.HZiZjZk[i,k,l])
        ans +=  (1.0/32)*(self.H-self.HZi[l]-self.HZi[k]+self.HZiZj[k,l]-self.ZiH[i]+self.ZiHZj[i,l]+self.ZiHZj[i,k]-self.ZiHZjZk[i,k,l])

        ans = 2*ans.real
        return ans

    def Hess_xx(self, i, j, k, l):
        ans = 0.0
        ans += -(1.0/64)*(self.H-self.HZi[i]-self.HZi[j]-self.HZi[k]-self.HZi[l]+self.HZiZj[i,j]+self.HZiZj[i,k]+self.HZiZj[i,l]+self.HZiZj[j,k]+self.HZiZj[j,l]+self.HZiZj[k,l]-self.HZiZjZk[i,j,k]-self.HZiZjZk[i,j,l]-self.HZiZjZk[i,k,l]-self.HZiZjZk[j,k,l]+self.HZiZjZkZl[i,j,k,l])
        ans +=  (1.0/64)*(self.H-self.ZiH[i]-self.ZiH[j]-self.HZi[k]-self.HZi[l]+self.ZiZjH[i,j]+self.ZiHZj[i,k]+self.ZiHZj[i,l]+self.ZiHZj[j,k]+self.ZiHZj[j,l]+self.HZiZj[k,l]-self.ZiZjHZk[i,j,k]-self.ZiZjHZk[i,j,l]-self.ZiHZjZk[i,k,l]-self.ZiHZjZk[j,k,l]+self.ZiZjHZkZl[i,j,k,l])

        ans = 2*ans.real
        return ans
