import sys
import time
import copy
import numpy as np
from scipy import optimize

from qulacs import QuantumState,PauliOperator,Observable,GeneralQuantumOperator
from qulacs.gate import X, Z,RY,Measurement,to_matrix_gate
from qulacs import gate
from qulacs.quantum_operator import create_quantum_operator_from_openfermion_file

from openfermion.chem import MolecularData
from openfermionpyscf import run_pyscf
from openfermion.transforms import get_fermion_operator, jordan_wigner

# our own modules
from algorithm1_emuqpe import *
from expectation import all_expectation

np.random.seed(seed=1000)

n = 8 # defined as 'n_v' in the paper

a = np.random.rand(n); a2 = np.random.rand(n)
a -= a2
W = np.random.rand(n*n); W2 = np.random.rand(n*n)
W -= W2; W = np.resize(W, (n,n))

c = np.random.rand(n); c2 = np.random.rand(n)
c -= c2
X = np.random.rand(n*n); X2 = np.random.rand(n*n)
X -= X2; X = np.resize(X, (n,n))

# a = np.loadtxt('b_param.txt')
# W = np.loadtxt('w_param.txt'); W = np.resize(W, (n,n))
# c = np.loadtxt('c_param.txt')
# X = np.loadtxt('x_param.txt'); X = np.resize(X, (n,n))
a *= 0.001
W *= 0.001
c *= 0.1
X *= 0.1

def E_exp(state):
    Hami = GeneralQuantumOperator(state.get_qubit_count())
    for i in range(Hamiltonian.get_term_count()):
        Hami.add_operator(Hamiltonian.get_term(i))
    Energy = Hami.get_expectation_value(state)
    return Energy

def function(a, W, c, X):
    m = significant_dig
    state = QuantumState(n)
    state = Sample(state, a, W)
    state = Phase(state, c, X)
    E_loc = E_exp(state).real
    return E_loc

def resize_param(x):
    ind = 0
    a = x[ind:n]; ind += n
    W = x[ind:ind+n*n]; ind += n*n; W = np.resize(W, (n,n))
    c = x[ind:ind+n]; ind += n
    X = x[ind:ind+n*n]; X = np.resize(X, (n,n))
    return [a, W, c, X]

def func(x):
    a, W, c, X = resize_param(x)
    Energy = function(a, W, c, X)
    print("Energy:", Energy)
    return Energy

def func_phase(x):
    c = x[:n]; X = x[n:]; X = np.resize(X, (n,n))
    Energy = function(a, W, c, X)
    print("Energy",Energy)
    return Energy

# ** IMPORTANT SUBROUTINES **
# Return the gradients of the energy with respect to the network parameters
# - fprime
# - fprime_phase
# - fprime_phase_SR
# - fprime_SR

def fprime(x):
    a, W, c, X = resize_param(x)
    m = significant_dig
    state = QuantumState(n)
    state = Sample(state, a, W) # Algorithm 1 (amplitude section)
    state = Phase(state, c, X) # Algorithm 1 (phase section)

    tools = all_expectation(state, n, Hamiltonian, False, False)
    tools.initialize()
    tools.calculate_jacobian()
    jac = tools.jac
    return jac

def fprime_phase(x):
    c = x[:n]; X = x[n:]; X = np.resize(X, (n,n))

    m = significant_dig
    state = QuantumState(n)
    state = Sample(state, a, W) # Algorithm 1 (amplitude section)
    state = Phase(state, c, X) # Algorithm 1 (phase section)

    tools = all_expectation(state, n, Hamiltonian, False, False)
    tools.initialize()
    tools.calculate_jacobian()
    jac = tools.jac[n+n*n:]
    return jac

def fprime_phase_SR(x):
    c = x[:n]; X = x[n:]; X = np.resize(X, (n,n))

    m = significant_dig
    state = QuantumState(n)
    state = Sample(state, a, W) # Algorithm 1 (amplitude section)
    state = Phase(state, c, X) # Algorithm 1 (phase section)

    tools = all_expectation(state, n, Hamiltonian, False, True)
    tools.initialize()
    tools.calculate_jacobian()
    tools.calculate_Smat()
    tools.calculate_SRjac()
    return tools.SRjac[n+n*n:], tools.original_norm_tau

def fprime_SR(x):
    a, W, c, X = resize_param(x)

    m = significant_dig
    state = QuantumState(n)
    state = Sample(state, a, W) # Algorithm 1 (amplitude section)
    state = Phase(state, c, X) # Algorithm 1 (phase section)

    tools = all_expectation(state, n, Hamiltonian, False, True)
    tools.initialize()
    tools.calculate_jacobian()
    tools.calculate_Smat()
    tools.calculate_SRjac()

    return tools.SRjac, tools.original_norm_the, tools.original_norm_tau

Hamiltonian = create_quantum_operator_from_openfermion_file("./lmo_hamiltonian.dat")

print("c")
print(c)
print("X")
print(X)

X = X.flatten()

ini = np.concatenate([c, X])

for count in range(1, 101):
    st = time.time()
    print("count:", count)
    alpha = 0.5
    print("alpha:", alpha)

    jac, original_norm_tau = fprime_phase_SR(ini)
    norm = np.linalg.norm(jac)
    print("jac_norm:", norm)
    if norm > 1:
        jac = jac/norm
    norm = np.linalg.norm(jac)

    ini -= alpha * jac
    now_E = func_phase(ini)
    c = ini[:n]; X = ini[n:]; X = np.resize(X, (n, n))

    X = X.flatten()
    ini = np.concatenate([c, X])

    if original_norm_tau < 1e-8:
        break

c = ini[:n]; X = ini[n:]; X = np.resize(X, (n, n))
print("D")
print(0.5*((c[0]+c[1]+X[0,0]+X[1,1]+X[0,1]+X[1,0]) - (c[2]+c[3]+X[2,2]+X[3,3]+X[2,3]+X[3,2])))

print("c")
print(c)
print("X")
print(X)
# print("message...",res.message)
# print("func...", res.fun)
# print("jac...", res.jac)
# print("iter num...", res.nit)
print()

W = W.flatten()
X = X.flatten()
ini = np.concatenate([a, W, c, X])

optimal_a = copy.deepcopy(a); optimal_W = copy.deepcopy(W)
optimal_c = copy.deepcopy(c); optimal_X = copy.deepcopy(X)
min_E = now_E

for count in range(100, 1000):
    print("count:", count)
    #alpha = 1.0/np.sqrt(np.sqrt(count))
    if count < 300:
        alpha = 1.0
    elif count < 700:
        alpha = 0.5
    else:
        alpha = 0.1
    print("alpha:", alpha)

    jac, original_norm_the, original_norm_tau = fprime_SR(ini)

    jac_the = jac[:n+n*n]
    jac_tau = jac[n+n*n:]
    norm_the = np.linalg.norm(jac_the)
    print("jac_norm_the:", norm_the)
    norm_tau = np.linalg.norm(jac_tau)
    print("jac_norm_tau:", norm_tau)

    if np.linalg.norm(jac_the) > 1.0:
        jac_the = jac_the/np.linalg.norm(jac_the)
    if np.linalg.norm(jac_tau) > 1.0:
        jac_tau = jac_tau/np.linalg.norm(jac_tau)

    jac = np.concatenate([jac_the, jac_tau])

    ini -= jac*alpha
    now_E = func(ini)
    a, W, c, X = resize_param(ini)
    if now_E < min_E:
        optimal_a = copy.deepcopy(a); optimal_W = copy.deepcopy(W)
        optimal_c = copy.deepcopy(c); optimal_X = copy.deepcopy(X)
        min_E = now_E
    a, W, c, X = resize_param(ini)
    print("D")
    print(0.5*((c[0]+c[1]+X[0,0]+X[1,1]+X[0,1]+X[1,0]) - (c[2]+c[3]+X[2,2]+X[3,3]+X[2,3]+X[3,2])))
    W = W.flatten()
    X = X.flatten()
    print("a...",np.linalg.norm(a),"W...",np.linalg.norm(W), "c...",np.linalg.norm(c), "X..",np.linalg.norm(X))

    ini = np.concatenate([a, W, c, X])
    if (original_norm_the < 1e-8) and (original_norm_tau < 1e-8):
        break

# res = optimize.minimize(func, ini, method="BFGS", jac=fprime, tol=1e-8)
# sol = res.x
# Ene = res.fun
# jac = res.jac
# iternum = res.nit

optimal_W = optimal_W.flatten()
optimal_X = optimal_X.flatten()
ini = np.concatenate([optimal_a, optimal_W, optimal_c, optimal_X])
Ene = func(ini)
jac = fprime(ini)
a = ini[:n]
W = ini[n:n*n+n]; W = np.resize(W,(n,n))
c = ini[n+n*n:2*n+n*n]
X = ini[n*n+2*n:]; X = np.resize(X, (n, n))

#print("sol...", res.x)

print("a")
print(a)
print("W")
print(W)
print("c")
print(c)
print("X")
print(X)
#print("message...",res.message)
print("func...", Ene)
print("jac...", jac)
#print("iter num...", res.nit)

state = QuantumState(n)
state = Sample(state, a, W)
state = Phase(state, c, X)
print()
dH_a = jac[:n]; dH_w = jac[n:n+n*n]; dH_c = jac[n+n*n:2*n+n*n]; dH_x = jac[2*n+n*n:]
newfile = "BM2_transbuta"
with open(newfile,mode="w") as f:
    f.write("Energy\n")
    f.write(f"{Ene}\n")
    f.write("a:\n")
    f.write(f"{a}\n")
    f.write("W:\n")
    f.write(f"{W}\n")
    f.write("c:\n")
    f.write(f"{c}\n")
    f.write("X:\n")
    f.write(f"{X}\n")
    f.write("D: 0.5*((c[0]+c[1]+X[0,0]+X[1,1]+X[0,1]+X[1,0]) - (c[2]+c[3]+X[2,2]+X[3,3]+X[2,3]+X[3,2]))\n")
    f.write(f"{0.5*((c[0]+c[1]+X[0,0]+X[1,1]+X[0,1]+X[1,0]) - (c[2]+c[3]+X[2,2]+X[3,3]+X[2,3]+X[3,2]))}\n")

    f.write("dH_a:\n")
    f.write(f"{dH_a}\n")
    f.write("dH_w:\n")
    f.write(f"{dH_w}\n")
    f.write("dH_c:\n")
    f.write(f"{dH_c}\n")
    f.write("dH_x:\n")
    f.write(f"{dH_x}\n")
    f.write("Phi:\n")
    f.write(f"{state.get_vector()[:2**n]}\n")
    f.write("|Phi|^2:\n")
    f.write(f"{np.abs(state.get_vector()[:2**n])**2}\n")
